from frasco.ext import *
from frasco.upload import url_for_upload
from frasco.helpers import url_for
from frasco.utils import slugify
from flask import current_app, request
import sqlalchemy as sqla
import hashlib
import urllib
import math
import random
import base64
import requests


def svg_to_base64_data(svg):
    return 'data:image/svg+xml;base64,' + base64.b64encode(svg)


class UserAvatarModelMixin(object):
    avatar_filename = sqla.Column(sqla.String)

    @property
    def avatar_url(self):
        return url_for_avatar(self)


class FrascoUsersAvatars(Extension):
    name = "frasco_users_avatars"
    defaults = {"url": None,
                "avatar_size": 80,
                "add_flavatar_route": False,
                "try_gravatar": True,
                "force_gravatar": False,
                "gravatar_size": None,
                "gravatar_email_column": None,
                "gravatar_default": "mm",
                "force_flavatar": False,
                "flavatar_size": "100%",
                "flavatar_name_column": None,
                "flavatar_font_size": 80,
                "flavatar_text_dy": "0.32em",
                "flavatar_length": 1,
                "flavatar_text_color": "#ffffff",
                "flavatar_bg_colors": ["#5A8770", "#B2B7BB", "#6FA9AB", "#F5AF29", "#0088B9", "#F18636", "#D93A37", "#A6B12E", "#5C9BBC", "#F5888D", "#9A89B5", "#407887", "#9A89B5", "#5A8770", "#D33F33", "#A2B01F", "#F0B126", "#0087BF", "#F18636", "#0087BF", "#B2B7BB", "#72ACAE", "#9C8AB4", "#5A8770", "#EEB424", "#407887"]}

    def _init_app(self, app, state):
        app.add_template_global(url_for_avatar)

        def flavatar(name, bgcolorstr=None):
            if bgcolorstr is None:
                bgcolorstr = request.args.get('bgcolorstr')
            svg = generate_first_letter_avatar_svg(name, bgcolorstr, request.args.get('size'))
            return svg, 200, {
                'Content-Type': 'image/svg+xml',
                'Cache-Control': 'public, max-age=31536000'
            }

        @app.route('/avatar/<hash>/<name>')
        def avatar(hash, name):
            if state.options['try_gravatar']:
                size = state.options['gravatar_size'] or state.options["avatar_size"]
                try:
                    r = requests.get(url_for_gravatar(hash, size=size, default=404))
                    if r.status_code != 404:
                        return r.content, 200, {'Content-Type': r.headers['content-type']}
                except Exception:
                    pass
            return flavatar(name, hash)

        if state.options['add_flavatar_route']:
            app.add_url_rule('/flavatar/<name>.svg', 'flavatar', flavatar)
            app.add_url_rule('/flavatar/<name>/<bgcolorstr>.svg', 'flavatar', flavatar)


def url_for_avatar(user):
    state = get_extension_state('frasco_users_avatars')
    if getattr(user, 'avatar_filename', None):
        return url_for_upload(user.avatar_filename)

    hash = None
    username = getattr(user, state.options["flavatar_name_column"] or 'username', None)
    if username:
        if isinstance(username, unicode):
            username = username.encode('utf-8')
        username = slugify(username.lower())
        hash = hashlib.md5(username).hexdigest()

    email = getattr(user, state.options["gravatar_email_column"] or 'email', None)
    if email:
        if isinstance(email, unicode):
            email = email.encode('utf-8')
        hash = hashlib.md5(email.lower()).hexdigest()
        if not username:
            username = slugify(email.split('@')[0])

    if state.options["force_flavatar"] and username:
        if state.options['add_flavatar_route']:
            return url_for('flavatar', name=username, bgcolorstr=hash, _external=True)
        return svg_to_base64_data(generate_first_letter_avatar_svg(username, hash))
    if state.options["force_gravatar"] and email:
        return url_for_gravatar(email)
    if state.options['url'] and email:
        return state.options["url"].format(email=email, email_hash=hash, username=username)
    return url_for('avatar', hash=hash, name=username, _external=True)


def url_for_gravatar(email, size=None, default=None):
    state = get_extension_state('frasco_users_avatars')
    hash = hashlib.md5(email.lower()).hexdigest()
    params = {
        's': size or state.options['gravatar_size'] or state.options["avatar_size"],
        'd': default or state.options['gravatar_default']
    }
    return "https://www.gravatar.com/avatar/%s?%s" % (hash, urllib.urlencode({k: v for k, v in params.items() if v is not None}))


def generate_first_letter_avatar_svg(name, bgcolorstr=None, size=None):
    state = get_extension_state('frasco_users_avatars')
    size = size or state.options['flavatar_size'] or state.options["avatar_size"]
    if size and isinstance(size, int):
        size = "%spx" % size

    svg_tpl = ('<svg xmlns="http://www.w3.org/2000/svg" pointer-events="none" viewBox="0 0 100 100" '
            'width="%(w)s" height="%(h)s" style="background-color: %(bgcolor)s;">%(letter)s</svg>')

    char_svg_tpl = ('<text text-anchor="middle" y="50%%" x="50%%" dy="%(dy)s" '
                    'pointer-events="auto" fill="%(fgcolor)s" font-family="'
                    'HelveticaNeue-Light,Helvetica Neue Light,Helvetica Neue,Helvetica, Arial,Lucida Grande, sans-serif" '
                    'style="font-weight: 400; font-size: %(size)spx">%(char)s</text>')

    if not name:
        text = '?'
    else:
        text = name[0:min(state.options['flavatar_length'], len(name))]
    colors_len = len(state.options['flavatar_bg_colors'])
    if bgcolorstr:
        bgcolor = sum([ord(c) for c in bgcolorstr]) % colors_len
    elif ord(text[0]) < 65:
        bgcolor = random.randint(0, colors_len - 1)
    else:
        bgcolor = int(math.floor((ord(text[0]) - 65) % colors_len))

    return svg_tpl % {
        'bgcolor': state.options['flavatar_bg_colors'][bgcolor],
        'w': size,
        'h': size,
        'letter': char_svg_tpl % {
            'dy': state.options['flavatar_text_dy'],
            'fgcolor': state.options['flavatar_text_color'],
            'size': state.options['flavatar_font_size'],
            'char': text
        }
    }



from frasco.ext import *
from frasco.utils import import_class
from flask import send_from_directory
from flask.wrappers import Request
from werkzeug import FileStorage
from io import BytesIO
from tempfile import TemporaryFile
import uuid
import os

from .backend import StorageBackend, split_backend_from_filename
from .utils import *


def _get_file_stream(self, total_content_length, content_type, filename=None, content_length=None):
    if total_content_length > 1024 * 500:
        return TemporaryFile('wb+', dir=os.environ.get('FRASCO_UPLOAD_TMP_DIR'))
    return BytesIO()

Request._get_file_stream = _get_file_stream


class FrascoUploadError(ExtensionError):
    pass


class FrascoUploadState(ExtensionState):
    def __init__(self, *args):
        super(FrascoUploadState, self).__init__(*args)
        self.backends = {}

    def _get_backend_options(self, name):
        options = dict(self.options)
        options.pop('backends', None)
        options.pop('backend', None) # make sure this key does not exists
        backend = name
        if name in self.options['backends']:
            if 'backend' in self.options['backends'][name]:
                options = self._get_backend_options(self.options['backends'][name]['backend'])
                backend = options['backend'] # make sure to keep the deepest level backend to resolve aliases properly
            options.update(self.options['backends'][name])
        options['backend'] = backend
        return options

    def get_backend(self, name=None):
        if isinstance(name, StorageBackend):
            return name
        if name is None:
            name = self.options['default_backend']
        if name not in self.backends:
            options = self._get_backend_options(name)
            backend_class = import_class(options['backend'], StorageBackend, "frasco.upload.backends")
            self.backends[name] = backend_class(name, options)
        return self.backends[name]


class FrascoUpload(Extension):
    name = 'frasco_upload'
    state_class = FrascoUploadState
    defaults = {"default_backend": "local",
                "backends": {},
                "upload_dir": "uploads",
                "upload_url": "/uploads",
                "upload_tmp_dir": None,
                "uuid_prefixes": True,
                "uuid_prefix_path_separator": False,
                "keep_filenames": True,
                "subfolders": False}

    def _init_app(self, app, state):
        app.add_template_global(url_for_upload)
        app.add_template_global(format_file_size)

        def send_uploaded_file(filename):
            return send_from_directory(os.path.abspath(state.options["upload_dir"]), filename)
        app.add_url_rule(state.options["upload_url"] + "/<path:filename>",
                         endpoint="static_upload",
                         view_func=send_uploaded_file)


def save_uploaded_file(file, filename=None, backend=None, **kwargs):
    state = get_extension_state('frasco_upload')
    return_filename_with_backend = kwargs.pop('return_filename_with_backend', False) or backend is True
    if not isinstance(file, FileStorage):
        file = FileStorage(file)
    if not filename:
        filename = generate_filename(file.filename, **kwargs)
    if not backend or backend is True:
        backend, filename = split_backend_from_filename(filename)
    state.get_backend(backend).save(file, filename)
    if return_filename_with_backend:
        return "%s://%s" % (backend or state.options['default_backend'], filename)
    return filename


def upload_stream(stream, filename, target_filename=None, **kwargs):
    return save_uploaded_file(FileStorage(stream, filename), target_filename, **kwargs)


def upload_file(pathname, filename=None, **kwargs):
    with open(pathname, 'rb') as f:
        return upload_stream(f, filename or pathname, **kwargs)


def delete_uploaded_file(filename, backend=None, **kwargs):
    if not backend:
        backend, filename = split_backend_from_filename(filename)
    get_extension_state('frasco_upload').get_backend(backend).delete(filename, **kwargs)


def url_for_upload(filename, backend=None, **kwargs):
    if not backend:
        backend, filename = split_backend_from_filename(filename)
    return get_extension_state('frasco_upload').get_backend(backend).url_for(filename, **kwargs)


def open_uploaded_file(filename, backend=None, **kwargs):
    if not backend:
        backend, filename = split_backend_from_filename(filename)
    return get_extension_state('frasco_upload').get_backend(backend).open(filename, **kwargs)



from flask import current_app, abort
from frasco.ext import get_extension_state
from itsdangerous import URLSafeTimedSerializer, BadSignature
from werkzeug.local import LocalProxy


__all__ = ('user_token_serializer', 'generate_user_token', 'read_user_token', 'read_user_token_or_404')


def get_user_token_serializer():
    return URLSafeTimedSerializer(current_app.config['SECRET_KEY'])


user_token_serializer = LocalProxy(get_user_token_serializer)


def generate_user_token(user, salt=None):
    return user_token_serializer.dumps(str(user.id), salt=salt)


def read_user_token(token, salt=None, max_age=None):
    try:
        user_id = user_token_serializer.loads(token, salt=salt, max_age=max_age)
        return get_extension_state('frasco_users').Model.query.get(user_id)
    except BadSignature:
        return None


def read_user_token_or_404(*args, **kwargs):
    user = read_user_token(*args, **kwargs)
    if not user:
        abort(404)
    return user



from jinja2.ext import Extension
import re


class AngularCompatExtension(Extension):
    """Jinja extension that does the bare minimum into making angular templates
    parsable by Jinja so gettext strings can be extacted.
    Removes angular one-time binding indicators and javascript ternary operator.
    """
    special_chars_re = re.compile(r"'[^']*'|\"[^\"]+\"|(\{[^{]+\}|[?:!&|$=]{1,3})")
    replacements = {'!': ' not ', '$': '', '=': '=', '==': '==',
                    '===': '==', '!=': '!=', '!==': '!=', '&&': ' and ', '||': ' or '}

    def process_expression(self, source, start):
        p = start
        end = p
        while True:
            end = source.find('}}', p)
            m = self.special_chars_re.search(source, p, end)
            if not m:
                break
            if m.group(1) is None:
                p = m.end(0)
                continue
            if m.group(1).startswith('{'):
                repl = 'True'
            else:
                repl = self.replacements.get(m.group(1), ' or ')
            p = m.start(1) + len(repl)
            source = source[:m.start(1)] + repl + source[m.end(1):]
        return source, end + 2

    def preprocess(self, source, name, filename=None):
        source = source.replace('{{::', '{{')
        p = 0
        while True:
            p = source.find('{{', p)
            if p == -1:
                break
            source, p = self.process_expression(source, p + 2)
        return source
